# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.util import boundingBox
import numpy as np
import ctypes
import moleculekit.home
import os
from numba import cuda, jit
import numba
from math import sqrt, exp
from functools import lru_cache
import logging


logger = logging.getLogger(__name__)

_order = ('hydrophobic', 'aromatic', 'hbond_acceptor', 'hbond_donor', 'positive_ionizable',
          'negative_ionizable', 'metal', 'occupancies')
libdir = moleculekit.home.home(libDir=True)
occupancylib = ctypes.cdll.LoadLibrary(os.path.join(libdir, "occupancy_ext.so"))


def viewVoxelFeatures(features, centers, nvoxels, voxelsize=1, draw='wireframe'):
    from moleculekit.vmdgraphics import VMDIsosurface
    from moleculekit.vmdviewer import getCurrentViewer

    if not isinstance(voxelsize, list) and not isinstance(voxelsize, np.ndarray):
        voxelsize = [voxelsize, voxelsize, voxelsize]

    features = features.reshape(list(nvoxels) + [len(_order),])
    centers = centers.reshape(list(nvoxels) + [3,])
    voxelsize = np.array(voxelsize)
    loweredge = np.min(centers, axis=(0, 1, 2)) - (voxelsize / 2)

    for i, name in enumerate(_order):
        VMDIsosurface(features[..., i], loweredge, voxelsize, color=i, name=name, draw=draw)

def rotateCoordinates(coords, rotations, center):
    from moleculekit.util import rotationMatrix

    def rotate(coords, rotMat, center=(0, 0, 0)):
        newcoords = coords - center
        return np.dot(newcoords, np.transpose(rotMat)) + center

    rotation = list(rotations)
    matx = rotationMatrix([1, 0, 0], rotation[0])
    maty = rotationMatrix([0, 1, 0], rotation[1])
    matz = rotationMatrix([0, 0, 1], rotation[2])

    coords = coords.copy()
    coords = rotate(coords, matx, center=center)
    coords = rotate(coords, maty, center=center)
    coords = rotate(coords, matz, center=center)
    return coords

def _getChannelRadii(molelements):
    from moleculekit.periodictable import periodictable
    radii = np.array([periodictable[e].vdw_radius for e in molelements])
    return radii


@lru_cache(maxsize=10)  # Caches results of the calculation to not have to recalculate it for each molecule
def _getGridCenters(x, y, z, resolution):
    firstdim = np.repeat(np.arange(x) * resolution, y*z)
    seconddim = np.tile(np.repeat(np.arange(y) * resolution, z), x)
    thirddim = np.tile(np.arange(z) * resolution, x*y)
    combined = np.vstack((firstdim.T, seconddim.T, thirddim.T)).T.astype(np.float64)
    combined = combined.reshape([x, y, z, 3])
    return combined


def getChannels(mol, aromaticNitrogen=False, version=2, validitychecks=True):
    from moleculekit.smallmol.smallmol import SmallMol
    from moleculekit.molecule import Molecule
    if isinstance(mol, SmallMol):
            channels = _getPropertiesRDkit(mol)
    elif isinstance(mol, Molecule):
        if version == 1:
            channels = _getAtomtypePropertiesPDBQT(mol)
        elif version == 2:
            from moleculekit.tools.atomtyper import getFeatures, getPDBQTAtomTypesAndCharges
            mol.atomtype, mol.charge = getPDBQTAtomTypesAndCharges(mol, aromaticNitrogen=aromaticNitrogen, validitychecks=validitychecks)
            channels = getFeatures(mol)

    if channels.dtype == bool:
        # Calculate for each channel the atom sigmas
        sigmas = _getChannelRadii(mol.get('element'))
        channels = sigmas[:, np.newaxis] * channels.astype(float)
    return channels, mol


def getCenters(mol=None, buffer=0, boxsize=None, center=None, voxelsize=1):
    """ Get a set of centers for voxelization.

    Parameters
    ----------
    mol : Molecule or SmallMol object
        Use this together with the buffer option.
    buffer : float
        The buffer space to add to the bounding box. This adds zeros to the grid around the protein so that properties
        which are at the edge of the box can be found in the center of one. Should be usually set to localimagesize/2.
    boxsize : list
        If this argument is None, the function will calculate the bounding box of the molecule, add to that the `buffer` argument
        to expand the bounding box in all dimensions and discretize it into voxels of `voxelsize` xyz length. In this case, 
        the `center` argument is ignored.
        If a list of [x, y, z] lengths is given in Angstrom, it will create a box of those dimensions centered around the `center` 
        argument.
    center : list
        The [x, y, z] coordinates of the center. Use this only in combination with the `boxsize` argument.
    voxelsize : float
        The voxel size in A
    """
    if boxsize is None:
        # Calculate the bbox and the number of voxels
        [bb_min, bb_max] = boundingBox(mol)
        bb_min -= buffer
        bb_max += buffer
        nvoxels = np.ceil((bb_max - bb_min) / voxelsize).astype(int) + 1  # TODO: Why the +1?
    else:
        boxsize = np.array(boxsize)
        center = np.array(center)
        nvoxels = np.ceil(boxsize / voxelsize).astype(int)
        bb_min = center - (boxsize / 2)  
    # Calculate grid centers
    centers = _getGridCenters(*list(nvoxels), voxelsize) + bb_min
    centers = centers.reshape(np.prod(nvoxels), 3).copy()  # Copy otherwise C code reads it wrong
    return centers, nvoxels


def getVoxelDescriptors(mol, boxsize=None, voxelsize=1, buffer=0, center=None, usercenters=None, userchannels=None, 
                        usercoords=None, aromaticNitrogen=False, method='C', version=2, validitychecks=True):
    """ Calculate descriptors of atom properties for voxels in a grid bounding the Molecule object.

    Parameters
    ----------
    mol : Molecule or SmallMol object
        A molecule
    boxsize : list
        If this argument is None, the function will calculate the bounding box of the molecule, add to that the `buffer` argument
        to expand the bounding box in all dimensions and discretize it into voxels of `voxelsize` xyz length. In this case, 
        the `center` argument is ignored.
        If a list of [x, y, z] lengths is given in Angstrom, it will create a box of those dimensions centered around the `center` 
        argument.
    voxelsize : float
        The voxel size in A
    buffer : float
        The buffer space to add to the bounding box. This adds zeros to the grid around the protein so that properties
        which are at the edge of the box can be found in the center of one. Should be usually set to localimagesize/2.
    center : list
        The [x, y, z] coordinates of the center. Use this only in combination with the `boxsize` argument.
    usercenters : np.ndarray
        A 2D array specifying the centers of the voxels. If None is given, it will calculate the centers from the above options
    userchannels : np.ndarray
        A 2D array of size (mol.numAtoms, nchannels) where nchannels is the number of channels we want to have.
        Each column i then has True (or a float) in the rows of the atoms which belong to channel i and False (or 0) 
        otherwise. Such boolean arrays can be obtained for example by using mol.atomselect.
        If the array is boolean, each atom will get assigned its VdW radius. If the array is float, these floats will 
        be used as the corresponding atom radii. Make sure the numpy array is of dtype=bool if passing boolean values.
        If no channels are given, the default ('hydrophobic', 'aromatic', 'hbond_acceptor', 'hbond_donor', 
        'positive_ionizable', 'negative_ionizable', 'metal', 'occupancies') channels will be used.
    usercoords : np.ndarray
        A numpy array containing the molecule coordinates. You can use this with userchannels and usercenters instead 
        of passing a `mol` object (set it to None if that's the case). 
    validitychecks : bool
        Set to False to disable validity checks for atomtyping. This improves performance but only use it if you
        are sure you have properly prepared your molecules.

    Returns
    -------
    features : np.ndarray
        A 2D array of size (centers, channels) containing the effect of each channel in the voxel with that center. 
    centers : np.ndarray
        A list of the centers of all boxes
    N : np.ndarray
        Is returned only when no user centers are passed. It corresponds to the number of centers in each of the x,y,z
        dimensions
    method : str
        Voxel descriptors can be calculated either with our C implementation or CUDA or NUMBA implementations.

    Examples
    --------
    >>> mol = Molecule('3PTB')
    >>> mol.filter('protein')
    >>> features, centers, N = getVoxelDescriptors(mol, buffer=8)
    >>> # Features can be reshaped to a 4D array (3D for each grid center in xyz, 1D for the properties) like this:
    >>> features = features.reshape(N[0], N[1], N[2], features.shape[1])
    >>> # The user can provide his own centers
    >>> features, centers = getVoxelDescriptors(mol, usercenters=[[0, 0, 0], [16, 24, -5]])
    """
    channels = userchannels
    if channels is None:
        channels, mol = getChannels(mol, aromaticNitrogen, version, validitychecks)

    if channels.dtype == bool:
        # Calculate for each channel the atom sigmas
        sigmas = _getChannelRadii(mol.element)
        channels = sigmas[:, np.newaxis] * channels.astype(float)

    nvoxels = None
    centers = usercenters
    if centers is None:
        centers, nvoxels = getCenters(mol, buffer, boxsize, center, voxelsize)

    coords = usercoords
    if coords is None:
        coords = mol.get('coords')
    if coords.ndim == 3:
        if coords.shape[2] != 1:
            raise RuntimeError('Only a single set of coordinates should be passed for voxelixation. ' \
                               'Make sure your coordinates are either 3D with a last dim of 1 or 2D.')
        else:
            coords = coords[:, :, 0]

    # Calculate features
    if method.upper() == 'C':
        features = _getOccupancyC(coords, centers.copy(), channels)
    elif method.upper() == 'CUDA':
        features = _getOccupancyCUDA(coords, centers, channels)
    elif method.upper() == 'NUMBA':
        features = _getOccupancyNUMBA(coords, centers, channels, 5)

    if nvoxels is None:
        return features, centers
    else:
        return features, centers, nvoxels


def _getPropertiesRDkit(smallmol):
    """
    Returns ndarray of shape (n_atoms x n_properties) molecule atom types,
    according to the following definitions and order:
        0. Hydrophibic
        1. Aromatic
        2. Acceptor
        3. Donor
        4. - Ionizable
        5. + Ionizable
        6. Metal (empty)
        7. Occupancy (No hydrogens)
    """
    from moleculekit.smallmol.util import factory
    n_atoms = smallmol.numAtoms

    atom_mapping = {"Hydrophobe": 0,
                    "LumpedHydrophobe": 0,
                    "Aromatic": 1,
                    "Acceptor": 2,
                    "Donor": 3,
                    "PosIonizable": 4,
                    "NegIonizable": 5}

    feats = factory.GetFeaturesForMol(smallmol._mol)
    properties = np.zeros((n_atoms, 8), dtype=bool)

    for feat in feats:
        fam = feat.GetFamily()
        if fam not in atom_mapping:  # Non relevant property
            continue
        properties[feat.GetAtomIds(), atom_mapping[fam]] = 1

    # Occupancy, ignoring hydrogens.
    properties[:, 7] = smallmol.get('element') != 'H'
    return properties


def _getAtomtypePropertiesPDBQT(mol):
    """ Matches PDBQT atom types to specific properties
    ('hydrophobic', 'aromatic', 'hbond_acceptor', 'hbond_donor', 'positive_ionizable', 'negative_ionizable', 'metal')

    Parameters
    ----------
    mol :
        A Molecule object

    Returns
    -------
    properties : dict
        A dictionary of atom masks for the Molecule showing for each property (key) which atoms (value) belong to it.
    """
    """ AutoDock 4 atom types http://autodock.scripps.edu/faqs-help/faq/faqsection_view?section=Scientific%20Questions
    #        --     ----  -----  -------  --------  ---  ---  -  --  -- --
    atom_par H      2.00  0.020   0.0000   0.00051  0.0  0.0  0  -1  -1  3    # Non H-bonding Hydrogen
    atom_par HD     2.00  0.020   0.0000   0.00051  0.0  0.0  2  -1  -1  3    # Donor 1 H-bond Hydrogen
    atom_par HS     2.00  0.020   0.0000   0.00051  0.0  0.0  1  -1  -1  3    # Donor S Spherical Hydrogen
    atom_par C      4.00  0.150  33.5103  -0.00143  0.0  0.0  0  -1  -1  0    # Non H-bonding Aliphatic Carbon
    atom_par A      4.00  0.150  33.5103  -0.00052  0.0  0.0  0  -1  -1  0    # Non H-bonding Aromatic Carbon
    atom_par N      3.50  0.160  22.4493  -0.00162  0.0  0.0  0  -1  -1  1    # Non H-bonding Nitrogen
    atom_par NA     3.50  0.160  22.4493  -0.00162  1.9  5.0  4  -1  -1  1    # Acceptor 1 H-bond Nitrogen
    atom_par NS     3.50  0.160  22.4493  -0.00162  1.9  5.0  3  -1  -1  1    # Acceptor S Spherical Nitrogen
    atom_par OA     3.20  0.200  17.1573  -0.00251  1.9  5.0  5  -1  -1  2    # Acceptor 2 H-bonds Oxygen
    atom_par OS     3.20  0.200  17.1573  -0.00251  1.9  5.0  3  -1  -1  2    # Acceptor S Spherical Oxygen
    atom_par F      3.09  0.080  15.4480  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Fluorine
    atom_par Mg     1.30  0.875   1.5600  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Magnesium
    atom_par MG     1.30  0.875   1.5600  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Magnesium
    atom_par P      4.20  0.200  38.7924  -0.00110  0.0  0.0  0  -1  -1  5    # Non H-bonding Phosphorus
    atom_par SA     4.00  0.200  33.5103  -0.00214  2.5  1.0  5  -1  -1  6    # Acceptor 2 H-bonds Sulphur
    atom_par S      4.00  0.200  33.5103  -0.00214  0.0  0.0  0  -1  -1  6    # Non H-bonding Sulphur
    atom_par Cl     4.09  0.276  35.8235  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Chlorine
    atom_par CL     4.09  0.276  35.8235  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Chlorine
    atom_par Ca     1.98  0.550   2.7700  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Calcium
    atom_par CA     1.98  0.550   2.7700  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Calcium
    atom_par Mn     1.30  0.875   2.1400  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Manganese
    atom_par MN     1.30  0.875   2.1400  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Manganese
    atom_par Fe     1.30  0.010   1.8400  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Iron
    atom_par FE     1.30  0.010   1.8400  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Iron
    atom_par Zn     1.48  0.550   1.7000  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Zinc
    atom_par ZN     1.48  0.550   1.7000  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Zinc
    atom_par Br     4.33  0.389  42.5661  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Bromine
    atom_par BR     4.33  0.389  42.5661  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Bromine
    atom_par I      4.72  0.550  55.0585  -0.00110  0.0  0.0  0  -1  -1  4    # Non H-bonding Iodine
    """
    from collections import OrderedDict
    props = OrderedDict()
    atomtypes = np.array([el.upper() for el in mol.atomtype])

    props['hydrophobic'] = (atomtypes == 'C') | (atomtypes == 'A')
    props['aromatic'] = atomtypes == 'A'
    props['hbond_acceptor'] = (atomtypes == 'NA') | (atomtypes == 'NS') | (atomtypes == 'OA') | (atomtypes == 'OS') | (
    atomtypes == 'SA')
    props['hbond_donor'] = _findDonors(mol, mol._getBonds())
    props['positive_ionizable'] = mol.charge > 0
    props['negative_ionizable'] = mol.charge < 0
    props['metal'] = (atomtypes == 'MG') | (atomtypes == 'ZN') | (atomtypes == 'MN') | \
                     (atomtypes == 'CA') | (atomtypes == 'FE')
    props['occupancies'] = (atomtypes != 'H') & (atomtypes != 'HS') & (atomtypes != 'HD')

    channels = np.zeros((len(atomtypes), len(props)), dtype=bool)
    for i, p in enumerate(_order):
        channels[:, i] = props[p]
    return channels


def _findDonors(mol, bonds):
    """ Finds atoms connected to HD and HS atoms to find the heavy atom donor (O or N)

    Parameters
    ----------
    mol :
        A Molecule object
    bonds : np.ndarray
        An array of atom bonds

    Returns
    -------
    donors : np.ndarray
        Boolean array indicating the donor heavy atoms in Mol
    """
    donors = np.zeros(mol.numAtoms, dtype=bool)
    hydrogens = np.where((mol.element == 'HD') | (mol.element == 'HS'))[0]
    for h in hydrogens:
        partners = bonds[bonds[:, 0] == h, 1]
        partners = np.hstack((partners, bonds[bonds[:, 1] == h, 0]))
        for p in partners:
            if mol.name[p][0] == 'N' or mol.name[p][0] == 'O':
                donors[p] = True
    return donors


def _getOccupancyC(coords, centers, channelsigmas):
    """ Calls the C code to calculate the voxels values for each property."""
    centers = centers.astype(np.float64)
    channelsigmas = channelsigmas.astype(np.float64)

    nchannels = channelsigmas.shape[1]
    occus = np.zeros((centers.shape[0], nchannels))

    occupancylib.descriptor_ext(centers.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                       coords.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
                       channelsigmas.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                       occus.ctypes.data_as(ctypes.POINTER(ctypes.c_double)),
                       ctypes.c_int(occus.shape[0]),  # n of centers
                       ctypes.c_int(coords.shape[0]),  # n of atoms
                       ctypes.c_int(nchannels))  # n of channels
    return occus


@jit('float64[:,:](float32[:,:], float64[:,:], float64[:,:], float64)', nopython=True)
def _getOccupancyNUMBA(coords, centers, channelsigmas, trunc):
    ncenters = centers.shape[0]
    natoms = coords.shape[0]
    nchannels = channelsigmas.shape[1]
    trunc = trunc * trunc  # Since we calculate the d**2 we need to get trunc**2
    occus = np.zeros((ncenters, nchannels))
    for a in range(natoms):
        coo = coords[a, :]
        atomsigmas = channelsigmas[a, :]

        for c in range(ncenters):
            cent = centers[c, :]
            d = np.zeros(3)
            d[0] = coo[0] - cent[0]
            d[1] = coo[1] - cent[1]
            d[2] = coo[2] - cent[2]
            d2 = d[0] * d[0] + d[1] * d[1] + d[2] * d[2]
            if d2 < trunc:
                d1 = 1 / sqrt(d2)
                for h in range(nchannels):
                    if atomsigmas[h] == 0:
                        continue
                    x = atomsigmas[h] * d1
                    x3 = x * x * x
                    x12 = x3 * x3 * x3 * x3
                    value = 1 - exp(-x12)
                    occus[c, h] = max(occus[c, h], value)
    return occus


def _memsetArray(array, val=0, threadsperblock=256):
    from math import ceil
    totalelem = np.prod(array.shape)
    nblocks = ceil(totalelem / threadsperblock)
    _memsetArrayCUDAkernel[nblocks, threadsperblock](array.reshape(totalelem), val)

@cuda.jit
def _memsetArrayCUDAkernel(array, val):
    threadidx = (cuda.threadIdx.x + (cuda.blockDim.x * cuda.blockIdx.x))
    if threadidx >= array.shape[0]:
        return
    array[threadidx] = val

def _getOccupancyCUDA(coords, centers, channelsigmas, trunc=5, device=0, resD=None, asnumpy=True, threadsperblock=256):
    #cuda.select_device(device)
    if resD is None:
        resD = cuda.device_array((centers.shape[0], channelsigmas.shape[1]), dtype=np.float32)
    _memsetArray(resD, val=0)

    natomblocks = int(np.ceil(coords.shape[0] / threadsperblock))
    blockspergrid = (centers.shape[0], natomblocks)

    centers = cuda.to_device(centers)
    coords = cuda.to_device(coords)
    channelsigmas = cuda.to_device(channelsigmas)
    _getOccupancyCUDAkernel[blockspergrid, threadsperblock](resD, coords, centers, channelsigmas, trunc * trunc)

    if asnumpy:
        return resD.copy_to_host()

@cuda.jit
def _getOccupancyCUDAkernel(occus, coords, centers, channelsigmas, trunc):
    centeridx = cuda.blockIdx.x
    blockidx = cuda.blockIdx.y
    atomidx = (cuda.threadIdx.x + (cuda.blockDim.x * blockidx))

    if atomidx >= coords.shape[0] or centeridx >= centers.shape[0]:
        return

    # TODO: Can remove this. Barely any speedup
    centcoor = cuda.shared.array(shape=(3), dtype=numba.float32)
    centcoor[0] = centers[centeridx, 0]
    centcoor[1] = centers[centeridx, 1]
    centcoor[2] = centers[centeridx, 2]
    cuda.syncthreads()

    dx = coords[atomidx, 0] - centcoor[0]
    dy = coords[atomidx, 1] - centcoor[1]
    dz = coords[atomidx, 2] - centcoor[2]
    d2 = dx * dx + dy * dy + dz * dz
    if d2 >= trunc:
        return

    d1 = 1 / sqrt(d2)
    for h in range(channelsigmas.shape[1]):
        if channelsigmas[atomidx, h] == 0:
            continue
        x = channelsigmas[atomidx, h] * d1
        value = 1 - exp(-(x ** 12))
        cuda.atomic.max(occus, (centeridx, h), value)


import unittest
class _TestVoxel(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.home import home
        from moleculekit.molecule import Molecule
        from moleculekit.tools.preparation import proteinPrepare
        from moleculekit.tools.autosegment import autoSegment

        self.testf = os.path.join(home(), 'test-data', 'test-voxeldescriptors')
        mol = Molecule(os.path.join(self.testf, '3PTB.pdb'))
        mol.filter('protein')
        mol = autoSegment(mol, field='both')
        mol = proteinPrepare(mol)
        mol.bonds = mol._guessBonds()
        self.mol = mol

    def test_radii(self):
        sigmas = _getChannelRadii(self.mol.element)
        refsigmas = np.load(os.path.join(self.testf, '3PTB_sigmas.npy'))
        assert np.allclose(sigmas, refsigmas)

    def test_celecoxib(self):
        from moleculekit.smallmol.smallmol import SmallMol
        sm = SmallMol(os.path.join(self.testf, 'celecoxib.mol2'))
        features, centers, nvoxels = getVoxelDescriptors(sm, buffer=1, version=2)
        reffeatures, refcenters, refnvoxels = np.load(os.path.join(self.testf, 'celecoxib_voxres.npy'))
        assert np.allclose(features, reffeatures)
        assert np.array_equal(centers, refcenters)
        assert np.array_equal(nvoxels, refnvoxels)

    def test_ledipasvir(self):
        from moleculekit.smallmol.smallmol import SmallMol
        sm = SmallMol(os.path.join(self.testf, 'ledipasvir.mol2'))
        features, centers, nvoxels = getVoxelDescriptors(sm, buffer=1, version=2)
        reffeatures, refcenters, refnvoxels = np.load(os.path.join(self.testf, 'ledipasvir_voxres.npy'))
        assert np.allclose(features, reffeatures)
        assert np.array_equal(centers, refcenters)
        assert np.array_equal(nvoxels, refnvoxels)

    def test_old_voxelization(self):
        from moleculekit.molecule import Molecule
        mol = Molecule(os.path.join(self.testf, '3ptb.pdbqt'))
        mol.element[mol.element == 'CA'] = 'Ca'
        features, centers, nvoxels = getVoxelDescriptors(mol, buffer=8, voxelsize=1, version=1)
        reffeatures, refcenters, refnvoxels = np.load(os.path.join(self.testf, '3PTB_voxres_old.npy'))
        assert np.allclose(features, reffeatures)
        assert np.array_equal(centers, refcenters)
        assert np.array_equal(nvoxels, refnvoxels)

    def test_featC(self):
        features, centers, nvoxels = getVoxelDescriptors(self.mol, method='C', version=2)
        reffeatures, refcenters, refnvoxels = np.load(os.path.join(self.testf, '3PTB_voxres.npy'))
        assert np.allclose(features, reffeatures)
        assert np.array_equal(centers, refcenters)
        assert np.array_equal(nvoxels, refnvoxels)

    def test_featNUMBA(self):
        features, centers, nvoxels = getVoxelDescriptors(self.mol, method='NUMBA', version=2)
        reffeatures, refcenters, refnvoxels = np.load(os.path.join(self.testf, '3PTB_voxres.npy'))
        assert np.allclose(features, reffeatures)
        assert np.array_equal(centers, refcenters)
        assert np.array_equal(nvoxels, refnvoxels)

    def test_compare_c_numba(self):
        import numpy as np
        centers = np.load(os.path.join(self.testf, '3PTB_centers_inp.npy'))
        coords = np.load(os.path.join(self.testf, '3PTB_coords_inp.npy'))
        sigmas = np.load(os.path.join(self.testf, '3PTB_channels_inp.npy'))
        centers = centers[::10, :].copy()

        res_C = _getOccupancyC(coords, centers, sigmas)
        # res_cuda = _getOccupancyCUDA(coords, centers, sigmas, 5)
        res_numba = _getOccupancyNUMBA(coords, centers, sigmas, 5)

        # assert np.abs(res_C - res_cuda).max() < 1e-4
        assert np.abs(res_C - res_numba).max() < 1e-4

if __name__ == '__main__':
    unittest.main(verbosity=2)







# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os
from copy import deepcopy
import multiprocessing
import math
import numpy as np
from rdkit import Chem
from rdkit.Chem.rdchem import HybridizationType, BondType
from collections import defaultdict
from moleculekit.smallmol.util import openbabelConvert, _depictMol, depictMultipleMols, convertToString
import logging
logger = logging.getLogger(__name__)


_bondtypes_IdxToType = BondType.values
_bondtypes_TypeToString = defaultdict(lambda: 'un')
_bondtypes_TypeToString.update({BondType.SINGLE: '1',
                                BondType.DOUBLE: '2',
                                BondType.TRIPLE: '3',
                                BondType.AROMATIC: 'ar'})
_bondtypes_FullStringToType = {'SINGLE': BondType.SINGLE,
                           'DOUBLE': BondType.DOUBLE,
                           'TRIPLE': BondType.TRIPLE}
_bondtypes_StringToType = {val: key for key, val in _bondtypes_TypeToString.items()}

_hybridizations_IdxToType = HybridizationType.values
_hybridizations_StringToType = {'S': HybridizationType.S,
                                'SP': HybridizationType.SP,
                                'SP2': HybridizationType.SP2,
                                'SP3': HybridizationType.SP3}


class SmallMol(object):
    """
    Class to manipulate small molecule structures

    Parameters
    ----------
    mol: rdkit.Chem.rdchem.Mol or filename or smile or moleculekit.smallmol.smallmol.SmallMol
        (i) Rdkit molecule or (ii) Location of molecule file (".pdb"/".mol2") or (iii) a smile string or iv) another
        SmallMol object or v) moleculekit.molecule.Molecule object
    ignore_errors: bool
        If True, errors will not be raised.
    force_reading: bool
        If True, and the mol provided is not accepted, the molecule will be initially converted into sdf
    fixHs: bool
        If True, the missing hydrogens are assigned, the others are correctly assinged into the graph of the molecule
    removeHs: bool
        If True, remove the hydrogens

    Examples
    --------
    >>> import os
    >>> from moleculekit.home import home
    >>> SmallMol('CCO')  # doctest: +SKIP
    >>> SmallMol(os.path.join(home(dataDir='test-smallmol'), 'ligand.pdb'), fixHs=False, removeHs=True )  # doctest: +SKIP
    >>> sm = SmallMol(os.path.join(home(dataDir='test-smallmol'), 'benzamidine.mol2'))
    >>> print(sm)                                     # doctest: +ELLIPSIS
    SmallMol with 18 atoms and 1 conformers
    Atom field - bondtype
    Atom field - charge
    ...

    .. rubric:: Methods
    .. autoautosummary:: moleculekit.smallmol.smallmol.SmallMol
       :methods:
    .. rubric:: Attributes
    .. autoautosummary:: moleculekit.smallmol.smallmol.SmallMol
       :attributes:
    """

    _atom_fields = ['idx', 'name', 'charge', 'formalcharge', 'element',  'chiral', 'hybridization',
                    'neighbors', 'bondtype', 'coords', 'chiraltags']

    _mol_fields = ['ligname', '_mol']

    _dtypes = {'idx': np.int,
                    'name': object,
                    'charge': np.float32,
                    'formalcharge': np.int,
                    'element': object,
                    'chiral': object,
                    'hybridization': int,
                    'neighbors': object,
                    'bondtype': object,
                    'coords': np.float32,
                    'chiraltags': object,
                    'bonds': np.uint32
                    }



    def __init__(self, mol, ignore_errors=False, force_reading=False, fixHs=True, removeHs=False):
        self._frame = 0

        _mol = self._initializeMolObj(mol, force_reading, ignore_errors)

        if removeHs:
            _mol = Chem.RemoveHs(_mol)
        if fixHs:
            _mol = Chem.AddHs(_mol, addCoords=True)

        self._mol = _mol

    def _initializeMolObj(self, mol, force_reading, ignore_errors):
        """
        Read the input and tries to convert it into a rdkit.Chem.rdchem.Mol obj

        Parameters
        ----------
        mol: str or rdkit.Chem.rdchem.Mol or moleculekit.smallmol.smallmol.SmallMol
            i) rdkit.Chem.rdchem.Mol ii) The path to the pdb/mol2 to load iii) The smile string iv) SmallMol object
            v) moleculekit.molecule.nolecule.Molecule
        force_reading: bool
        If the mol provided is not accepted, the molecule will be initially converted into sdf

        Returns
        -------
        _mol: rdkit.Chem.Molecule object
            The rdkit molecule
        smallMol: moleculekit.smallmol.smallmol.SmallMol
            The smallMol object if SmallMol was passed
        """
        from moleculekit.molecule import Molecule

        _mol = None
        if isinstance(mol, Chem.Mol):
            _mol = mol
        elif isinstance(mol, Molecule):
            _mol = self._fromMolecule(mol)
        elif isinstance(mol, str):
            if os.path.isfile(mol):
                name_suffix = os.path.splitext(mol)[-1]
                # load mol2 file
                if name_suffix == ".mol2":
                    _mol = Chem.MolFromMol2File(mol, removeHs=False)
                # load pdb file
                elif name_suffix == ".pdb":
                    _mol = Chem.MolFromPDBFile(mol, removeHs=False)
                # if the file failed to be loaded and 'force_reading' = True, file convert to sdf and than loaded
                if _mol is None and force_reading:
                    logger.warning('Reading {} with force_reading procedure'.format(mol))
                    sdf = openbabelConvert(mol, name_suffix, 'sdf')
                    _mol = Chem.SDMolSupplier(sdf, removeHs=False)[0]
                    os.remove(sdf)
            # assuming is a smile
            # TODO validate it. Implement smarts recognition
            else:
                # try with smiles
                psmile = Chem.SmilesParserParams()
                psmile.removeHs = False
                _mol = Chem.MolFromSmiles(mol, psmile)

        if _mol is None and not ignore_errors:
            if isinstance(mol, str):
                frerr = ' Try by setting the force_reading option as True.' if not force_reading else ''
                raise ValueError('Failed to read file {}.{}'.format(mol, frerr))
            if isinstance(mol, Molecule):
                raise ValueError('Failed convering Molecule to SmallMol')
        return _mol

    @property
    def _idx(self):
        return np.array([a.GetIdx() for a in self._mol.GetAtoms()], dtype=SmallMol._dtypes['idx'])

    @property
    def _element(self):
        return np.array([a.GetSymbol() for a in self._mol.GetAtoms()], dtype=SmallMol._dtypes['element'])

    @property
    def _name(self):
        return np.array(['{}{}'.format(a.GetSymbol(), a.GetIdx()) for a in self._mol.GetAtoms()], dtype=SmallMol._dtypes['name'])

    @property
    def _charge(self):
        return np.array([a.GetPropsAsDict()['_TriposPartialCharge'] if a.HasProp('_TriposPartialCharge') else 0.0 for a in self._mol.GetAtoms()], dtype=SmallMol._dtypes['charge'])

    @property
    def _formalcharge(self):
        return np.array([a.GetFormalCharge() for a in self._mol.GetAtoms()], dtype=SmallMol._dtypes['formalcharge']) 

    @property
    def _chiral(self):
        return np.array([a.GetPropsAsDict()['_CIPCode'] if a.HasProp('_CIPCode') else '' for a in self._mol.GetAtoms()], dtype=SmallMol._dtypes['chiral'])  

    @property
    def _hybridization(self):
        return np.array([int(a.GetHybridization()) for a in self._mol.GetAtoms()], dtype=SmallMol._dtypes['hybridization'])

    @property
    def _neighbors(self):
        return np.array([[na.GetIdx() for na in a.GetNeighbors()] for a in self._mol.GetAtoms()], dtype=SmallMol._dtypes['neighbors']) 

    @property
    def _neighborsbondtype(self):
        return np.array([[self._mol.GetBondBetweenAtoms(int(a), int(n)).GetBondType() for n in neighs] for a, neighs in enumerate(self._neighbors)], dtype=SmallMol._dtypes['bondtype'])

    @property
    def _bonds(self):
        return np.vstack([[bo.GetBeginAtomIdx(), bo.GetEndAtomIdx()] for bo in self._mol.GetBonds()]).astype(SmallMol._dtypes['bonds'])

    @property
    def _bondtype(self):
        return np.array([_bondtypes_TypeToString[self._mol.GetBondBetweenAtoms(int(b1), int(b2)).GetBondType()] for b1, b2 in self._bonds], dtype=SmallMol._dtypes['bondtype'])

    @property
    def _chiraltags(self):
        return np.array([a.GetChiralTag() for a in self._mol.GetAtoms()], dtype=SmallMol._dtypes['chiraltags'])

    @property
    def ligname(self):
        return self._mol.GetProp('_Name') if self._mol.HasProp('_Name') else 'UNK'

    @property
    def _resname(self):
        return np.array([self.ligname.strip() for _ in range(self.numAtoms)], dtype=object)

    @property
    def _coords(self):
        if self._mol.GetNumConformers() != 0:
            return np.array(np.stack([cc.GetPositions() for cc in self._mol.GetConformers()], axis=2), dtype=SmallMol._dtypes['coords'])
        else:
            return np.zeros((self.numAtoms, 3, 0), dtype=SmallMol._dtypes['coords'])

    @property
    def _atomtype(self):
        return np.array([a.GetPropsAsDict()['_TriposAtomType'] if a.HasProp('_TriposAtomType') else '' for a in self._mol.GetAtoms()], dtype=object)

    @property
    def frame(self):
        if self._frame < 0 or self._frame >= self.numFrames:
            raise NameError("frame out of range")
        return self._frame

    @frame.setter
    def frame(self, value):
        if value < 0 or ((self.numFrames != 0) and (value >= self.numFrames)):
            raise NameError("Frame index out of range. Molecule contains {} frame(s). Frames are 0-indexed.".format(self.numFrames))
        self._frame = value

    # Dummy properties for compatibility with Molecule
    @property
    def _serial(self):
        return np.arange(self.numAtoms, dtype=np.int32)

    @property
    def _resid(self):
        return np.ones(self.numAtoms, dtype=np.int32)

    @property
    def _occupancy(self):
        return np.zeros(self.numAtoms, dtype=np.float32)

    @property
    def _beta(self):
        return np.zeros(self.numAtoms, dtype=np.float32)

    @property
    def _masses(self):
        return np.zeros(self.numAtoms, dtype=np.float32)

    @property
    def _insertion(self):
        return np.array(['' for _ in range(self.numAtoms)], dtype=object)

    @property
    def _chain(self):
        return np.array(['' for _ in range(self.numAtoms)], dtype=object)

    @property
    def _segid(self):
        return np.array(['' for _ in range(self.numAtoms)], dtype=object)

    @property
    def _altloc(self):
        return np.array(['' for _ in range(self.numAtoms)], dtype=object)

    @property
    def _record(self):
        return np.array(['HETATM' for _ in range(self.numAtoms)], dtype=object)

    @property
    def numAtoms(self):
        return self._mol.GetNumAtoms()

    @property
    def numFrames(self):
        return self._mol.GetNumConformers()

    @property
    def _totalcharge(self):
        return sum(self._formalcharge)

    def _getBonds(self, fileBonds=True, guessBonds=True):
        return self._bonds

    def filter(self, sel):
        # Not implemented
        raise NotImplementedError('Filtering atoms not supported yet.')

    def copy(self):
        """
        Create a copy of the molecule object

        Returns
        -------
        newsmallmol : :class:`SmallMol`
            A copy of the object
        """
        return SmallMol(Chem.Mol(self._mol))  # Chem.Mol creates a deep copy of the C++ object

    def get(self, returnField, sel='all', convertType=True, invert=False):
        """
        Returns the property for the atom specified with the selection. The selection is another atom property

        Parameters
        ----------
        returnField: str
            The field of the atom to return
        sel: str
            The selection string. atom field name followed by spaced values for that field
        convertType: bool
            If True, and where possible the returnField is converted in rdkit object
            Default: True
        invert: bool
            If True, the selection is inverted
            Default: False

        Returns
        -------
        values: np.array
            The array of values for the property

        Example
        -------
        >>> sm.get('element', 'idx 0 1 7')  # doctest: +SKIP
        array(['C', 'C', 'H'],
              dtype='<U1')
        >>> sm.get('hybridization', 'element N')  # doctest: +SKIP
        array([rdkit.Chem.rdchem.HybridizationType.SP2,
               rdkit.Chem.rdchem.HybridizationType.SP2], dtype=object)
        >>> sm.get('hybridization', 'element N', convertType=False)
        array([3, 3])
        >>> sm.get('element', 'hybridization sp2')  # doctest: +SKIP
        array(['C', 'C', 'C', 'C', 'C', 'C', 'C', 'N', 'N'],
              dtype='<U1')
        >>> sm.get('element', 'hybridization S')  # doctest: +SKIP
        array(['H', 'H', 'H', 'H', 'H', 'H', 'H', 'H', 'H'],
              dtype='<U1')
        >>> sm.get('element', 'hybridization 1')  # doctest: +SKIP
        array(['H', 'H', 'H', 'H', 'H', 'H', 'H', 'H', 'H'],
              dtype='<U1')
        >>> sm.get('atomobject', 'element N')  # doctest: +SKIP
        array([<rdkit.Chem.rdchem.Atom object at 0x7faf616dd120>,
               <rdkit.Chem.rdchem.Atom object at 0x7faf616dd170>], dtype=object)
        """
        if sel == 'all':
            sel = 'idx {}'.format(convertToString(self._idx.tolist()))
        # get the field key and the value to grep
        key = sel.split()[0]
        selector = sel.split()[1:]

        if key not in self._atom_fields:
            raise KeyError('The property passed {} does not exist'.format(key))
        if len(selector) == 0:
            raise ValueError('No selection was provided')

        # get the returnField and process exceptional field
        if hasattr(self, '_'+key):
            _arrayFrom = self.__getattribute__('_'+key)
        else:
            _arrayFrom = self.__getattribute__(key)

        # special selector for hybridization: can be idx, or rdkit.Chem.rdchem.HybridizationType
        if key == 'hybridization':
            try:
                selector = [_hybridizations_StringToType[s.upper()] for s in selector]
            except:
                pass

        _dtype = self._dtypes[key]
        if _dtype is not object:
            selector = [_dtype(s) for s in selector]
        idxs = np.concatenate([np.where(_arrayFrom == s)[0] for s in selector])
        if invert:
            idxs = np.array([i for i in self._idx if i not in idxs])
        idxs = np.sort(idxs)

        if returnField == 'atomobject':
            return self.getAtoms()[idxs]
        elif returnField == 'bondtype':
            if convertType:
                return np.array([[_bondtypes_IdxToType[bt] for bt in neighbt] for neighbt in self._neighborsbondtype[idxs]], dtype=object)
            else:
                return self._neighborsbondtype[idxs]
        elif returnField == 'hybridization' and convertType:
            _arrayTo = np.array([_hybridizations_IdxToType[v] for v in self._hybridization], dtype=object)
        else:
            _arrayTo = self.__getattribute__('_'+returnField)

        return _arrayTo[idxs]

    def foundBondBetween(self, sel1, sel2, bondtype=None):
        """
        Returns True if at least a bond is found between the two selections. It is possible to check for specific bond
        type. A tuple is returned in the form (bool, [ [(idx1,idx2), rdkit.Chem.rdchem.BondType]] ])

        Parameters
        ----------
        sel1: str
            The selection for the first set of atoms
        sel2: str
            The selection for the second set of atoms
        bondtype: str or int
            The bondtype as index or string
            Default: None

        Returns
        -------
        isbond: bool
            True if a bond was found
        details: list
            A list of lists with the index of atoms in the bond and its type
        """

        if isinstance(bondtype, str):
            _btype = _bondtypes_FullStringToType[bondtype]
        else:
            _btype = bondtype

        atomIdx_sel1 = self.get('idx', sel1)
        neighbors_sel1 = self.get('neighbors', sel1)
        bondtypes_sel1 = self.get('bondtype', sel1)
        atomIdx_sel2 = self.get('idx', sel2)

        founds = []
        for aIdx, neighbors_set, btype_set in zip(atomIdx_sel1, neighbors_sel1, bondtypes_sel1):
            for neighbor, btype in zip(neighbors_set, btype_set):
                if neighbor in atomIdx_sel2:
                    btype_str = _bondtypes_TypeToString[btype]
                    if _btype is not None and _btype == btype:
                        founds.append([(aIdx, neighbor), btype_str])
                    elif bondtype is None:
                        founds.append([(aIdx, neighbor), btype_str])
        if len(founds) != 0:
            return True, founds

        return False

    def isChiral(self, returnDetails=False):
        """
        Returns True if the molecule has at least one chiral atom. If returnDetails is set as True,
        a list of tuples with the atom idx and chiral type is returned.

        Parameters
        ----------
        returnDetails: bool
            If True, returns the chiral atoms and their chiral types
            Default: False

        Returns
        -------
        ischiral: bool
            True if the atom has at least a chiral atom
        details: list
            A list of tuple with the chiral atoms and their types

        Example
        -------
        >>> chiralmol.isChiral()  # doctest: +SKIP
        True
        >>> chiralmol.isChiral(returnDetails=True)  # doctest: +SKIP
        (True, [('C2', 'R')])
        """

        _chirals = self._chiral

        idxs = np.where(_chirals != '')[0]
        if len(idxs) == 0:
            return False
        if returnDetails:
            idxs = idxs.astype(str)
            idxs_str = ' '.join(idxs)
            names = self.get('name', 'idx {}'.format(idxs_str))
            chirals = self.get('chiral', 'idx {}'.format(idxs_str))

            return True, [(a, c) for a, c in zip(names, chirals)]

        return True

    def getAtoms(self):
        """
        Retuns an array with the rdkit.Chem.rdchem.Atom present in the molecule
        """
        return np.array([a for a in self._mol.GetAtoms()])

    def getCenter(self):
        """
        Returns geometrical center of molecule conformation
        """
        coords = self._coords[:, :, self.frame]
        return coords.mean(axis=0).astype(np.float32)

    def generateConformers(self, num_confs=400,  optimizemode='mmff', align=True, append=True):
        """
        Generates ligand conformers

        Parameters
        ----------
        num_confs: int
           Number of conformers to generate.
        optimizemode: str
            The optimizemode to use. Can be  'uff', 'mmff'
            Default: 'mmff'
        align: bool
            If True, the conformer are aligned to the first one
            Default: True
        append: bool
            If False, the current conformers are deleted
            Default: True

        """
        from rdkit.Chem.AllChem import UFFOptimizeMolecule, MMFFOptimizeMolecule, EmbedMultipleConfs
        from rdkit.Chem.rdMolAlign import AlignMolConformers

        if not append:
            self.dropFrames(np.arange(self.numFrames))

        # get the rdkit mol and copied it.
        mol = deepcopy(self._mol)
        # hydrogens are added for safety
        mol = Chem.AddHs(mol)

        # generating conformations
        ids = EmbedMultipleConfs(mol, clearConfs=False, numConfs=num_confs, pruneRmsThresh=1., maxAttempts=10000)
        if optimizemode not in ['uff', 'mmff']:
            raise ValueError('Unknown optimizemode. Should be  "uff", "mmff"')
        # optimizing conformations depends on the optimizemode passed
        for id in ids:
            if optimizemode == 'mmff':
                MMFFOptimizeMolecule(mol, confId=id)
            elif optimizemode == 'uff':
                UFFOptimizeMolecule(mol, confId=id)

        if align:
            AlignMolConformers(mol)

        self._mol = mol

    def align(self, refmol):
        from rdkit.Chem.rdMolAlign import GetO3A
        from moleculekit.molecule import Molecule

        if isinstance(refmol, SmallMol):
            refmol = refmol._mol
        if isinstance(refmol, Molecule):
            refmol = SmallMol(refmol)._mol

        pyO3A = GetO3A(self._mol, refmol)
        rmsd = pyO3A.Align()
        logger.info('Alignment with a RMSD of {}'.format(rmsd))

    def dropFrames(self, frames='all'):
        if isinstance(frames, int):
            frames = np.array([frames,])
        if isinstance(frames, list):
            frames = np.array(frames)
        if frames == 'all':
            ids = [cc.GetId() for cc in self._mol.GetConformers()]
            for f in ids:
                self._mol.RemoveConformer(int(f))
        else:
            if np.any(frames >= self.numFrames):
                raise RuntimeError('Cannot remove more frames than existing which is {}'.format(self.numFrames))
            ids = [cc.GetId() for cc in self._mol.GetConformers()]
            for f in frames:
                self._mol.RemoveConformer(int(ids[f]))

    def write(self, fname, frames=None, merge=True):
        ext = os.path.splitext(fname)[-1]
        if ext == '.sdf':
            chemwrite = Chem.SDWriter
            # If merge is set as True a unique file is generated
            if merge:
                writer = chemwrite(fname)
                writer.write(self._mol)
            else: # If merge is set as False a file is created for each conformer
                if frames is None:
                    frames = list(range(self.numFrames))
                os.makedirs(os.path.dirname(fname), exist_ok=True)
                for i in frames:
                    fname_first = os.path.splitext(fname)[0]
                    currfname = os.path.join('{}_{}{}'.format(fname_first, i, ext))
                    writer = chemwrite(currfname)
                    writer.write(self._mol, confId=i)
        else:
            mol = self.toMolecule()
            mol.write(fname, frames)

    def view(self, *args, **kwargs):
        self.toMolecule().view(*args, **kwargs)

    def toSMARTS(self, explicitHs=False):
        """
        Returns the smarts string of the molecule

        Parameters
        ----------
        explicitHs: bool
            Set as True for keep the hydrogens

        Returns
        -------
        smart: str
            The smarts string
        """
        from copy import deepcopy
        rmol = deepcopy(self._mol)
        if not explicitHs and len(np.where(self._element == 'H')[0]) != 0:
            rmol = Chem.RemoveHs(rmol)

        return Chem.MolToSmarts(rmol, isomericSmiles=True)

    def toSMILES(self, explicitHs=False, kekulizeSmile=True):
        """
        Returns the smiles string of the molecule

        Parameters
        ----------
        explicitHs: bool
            Set as True to keep the hydrogens
        kekulizeSmile: bool
            Set as True to return the kekule smile format

        Returns
        -------
        smi: str
            The smiles string
        """
        from copy import deepcopy
        rmol = deepcopy(self._mol)

        if not explicitHs and len(np.where(self._element == 'H')[0]) != 0:
            rmol = Chem.RemoveHs(rmol)

        if kekulizeSmile:
            Chem.Kekulize(rmol)
            smi = Chem.MolToSmiles(rmol, isomericSmiles=True, kekuleSmiles=True)
        else:
            smi = Chem.MolToSmiles(rmol, isomericSmiles=True)

        return smi

    def _fromMolecule(self, mol):
        """
        Returns the rdkit.Chem.rdchem.Mol object from an moleculekit.molecule.Molecule one

        Parameters
        ----------
        mol: moleculekit.molecule.Molecule
            The moleculekit Molecule object

        Returns
        -------
        rmol: rdkit.Chem.rdchem.Mol
            The rdkit molecule object
        """

        from tempfile import NamedTemporaryFile

        tmpmol2 = NamedTemporaryFile(suffix='.mol2').name
        mol.write(tmpmol2)

        return Chem.MolFromMol2File(tmpmol2, removeHs=False)

    def toMolecule(self, formalcharges=False, ids=None):
        """
        Return the moleculekit.molecule.Molecule

        Parameters
        ----------
        formalcharges: bool
            If True,the formal charges are used instead of partial ones
        ids: list
            The list of conformer ids to store in the moleculekit Molecule object- If None, all are returned
            Default: None

        Returns
        -------
        mol: moleculekit.molecule.Molecule
            The moleculekit Molecule object

        """
        from moleculekit.molecule import Molecule

        class NoConformerError(Exception):
            pass

        if ids is None:
            ids = np.arange(self.numFrames)

        if self.numFrames == 0:
            raise NoConformerError("No Conformers are found in the molecule. Generate at least one conformer.")
        elif not isinstance(ids, list) and not isinstance(ids, np.ndarray):
            raise ValueError('The argument ids should be a list of confomer ids')

        mol = Molecule()
        mol.empty(self.numAtoms)
        mol.record[:] = 'HETATM'
        mol.resname[:] = self.ligname[:3]
        mol.resid[:] = self._resid
        mol.coords = self._coords[:, :, ids]
        mol.name[:] = self._name
        mol.element[:] = self._element
        if formalcharges:
            mol.charge[:] = self._formalcharge
        else:
            mol.charge[:] = self._charge
        mol.box = np.zeros((3, self.numFrames), dtype=np.float32)
        mol.viewname = self.ligname
        mol.bonds = self._bonds
        mol.bondtype = self._bondtype
        mol.atomtype = self._atomtype
        return mol

    def depict(self, sketch=True, filename=None, ipython=False, optimize=False, optimizemode='std', removeHs=True,
               atomlabels=None, highlightAtoms=None, resolution=(400, 200)):
        """
        Depicts the molecules. It is possible to save it into an svg file and also generates a jupiter-notebook rendering

        Parameters
        ----------
        sketch: bool
            Set to True for 2D depiction
        filename: str
            Set the filename for the svg file
        ipython: bool
            Set to True to return the jupiter-notebook rendering
        optimize: bool
            Set to True to optimize the conformation. Works only with 3D.
        optimizemode: ['std', 'mmff']
            Set the optimization mode for 3D conformation
        removeHs: bool
            Set to True to hide hydrogens in the depiction
        atomlabels: str
            Accept any combinations of the following pararemters as unique string '%a%i%c%*' a:atom name, i:atom index,
            c:atom formal charge (+/-), *:chiral (* if atom is chiral)
        highlightAtoms: list
            List of atom to highlight. It can be also a list of atom list, in this case different colors will be used
        resolution: tuple of integers
            Resolution in pixels: (X, Y)

        Returns
        -------
            ipython_svg: SVG object if ipython is set to True

        Example
        -------
        >>> sm.depict(ipython=True, optimize=True, optimizemode='std')  # doctest: +SKIP
        >>> sm.depict(ipython=True, sketch=True)  # doctest: +SKIP
        >>> sm.depict(ipython=True, sketch=True)  # doctest: +SKIP
        >>> sm.depict(ipython=True, sketch=True, atomlabels="%a%i%c")  # doctest: +SKIP
        >>> ids = np.intersect1d(sm.get('idx', 'hybridization SP2'), sm.get('idx', 'element C'))  # doctest: +SKIP
        >>> sm.depict(ipython=True, sketch=True,highlightAtoms=ids.tolist(), removeHs=False)  # doctest: +SKIP
        """
        from rdkit import Chem
        from rdkit.Chem.AllChem import Compute2DCoords, EmbedMolecule, MMFFOptimizeMolecule, ETKDG
        from copy import deepcopy

        if sketch and optimize:
            raise ValueError('Impossible to use optmization in  2D sketch representation')

        if optimizemode not in ['std', 'mmff']:
            raise ValueError('Optimization mode {} not understood. Can be "std" or "ff"'.format(optimizemode))

        elements = self._element
        indexes = self._idx
        formalcharges = self._formalcharge
        chirals = self._chiral

        _mol = deepcopy(self._mol)


        if sketch:
            Compute2DCoords(_mol)

        if removeHs:
            _mol = Chem.RemoveHs(_mol)
            elements = self.get('element', 'element H', invert=True)
            indexes = self.get('idx', 'element H', invert=True)
            formalcharges = self.get('formalcharge', 'element H', invert=True)
            chirals = self.get('chiral', 'element H', invert=True)

        _labelsFunc = ['a', 'i', 'c', '*']

        if atomlabels is not None:
            labels = atomlabels.split('%')[1:]
            formalcharges = ['' if c == 0 else "+" if c == 1 else "-" for c in formalcharges]
            chirals = ['' if c == '' else '*' for c in chirals]
            values = [elements, indexes, formalcharges, chirals]

            idxs = [_labelsFunc.index(l) for l in labels]
            labels_required = [values[i] for i in idxs]
            atomlabels = ["".join([str(i) for i in a]) for a in list(zip(*labels_required))]

        if optimize:
            if optimizemode == 'std':
                EmbedMolecule(_mol, ETKDG())
            elif optimizemode == 'mmff':
                MMFFOptimizeMolecule(_mol)

        return _depictMol(_mol, filename=filename, ipython=ipython,  atomlabels=atomlabels,
                          highlightAtoms=highlightAtoms, resolution=resolution)

    def __repr__(self):
        return '<{}.{} object at {}>\n'.format(self.__class__.__module__, self.__class__.__name__, hex(id(self))) \
               + self.__str__()

    def __str__(self):
        def formatstr(name, field):
            if isinstance(field, np.ndarray) or isinstance(field, list):
                rep = '{} shape: {}'.format(name, np.shape(field))
            elif field == 'reps':
                rep = '{}: {}'.format(name, len(self.reps.replist))
            else:
                rep = '{}: {}'.format(name, field)
            return rep

        rep = 'SmallMol with {} atoms and {} conformers'.format(self.numAtoms, self.numFrames)
        for p in sorted(self._atom_fields):
            if p.startswith('_'):
                continue
            rep += '\n'
            rep += 'Atom field - {}'.format(p)
        for j in sorted(self.__dict__.keys() - list(SmallMol._atom_fields)):
            if j[0] == '_':
                continue
            rep += '\n'
            rep += formatstr(j, self.__dict__[j])

        return rep

if __name__ == '__main__':

    import doctest
    import os
    from moleculekit.home import home
    from moleculekit.smallmol.smallmollib import SmallMolLib

    lib = SmallMolLib(os.path.join(home(dataDir='test-smallmol'), 'fda_drugs_light.sdf'))
    sm = SmallMol(os.path.join(home(dataDir='test-smallmol'), 'benzamidine.mol2'))
    doctest.testmod(extraglobs={'lib': lib.copy(), 'sm': sm.copy()})




import setuptools
import unittest

def my_test_suite():
    test_loader = unittest.TestLoader()
    test_suite = test_loader.discover(start_dir='moleculekit', pattern='*.py')
    return test_suite

if __name__ == '__main__':
    with open('README.md', 'r') as fh:
        long_description = fh.read()

    setuptools.setup(
        name='moleculekit',
        version='0.1.6',
        author='Acellera',
        author_email='info@acellera.com',
        description='A molecule reading/writing and manipulation package.',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/acellera/moleculekit/',
        classifiers=[
            'Programming Language :: Python :: 3.6',
            'Operating System :: POSIX :: Linux',
        ],

        packages=setuptools.find_packages(include=['moleculekit*'], exclude=['test-data', '*test*', ]),

        package_data={
            'moleculekit': ['lib/*/*', 'vmd_wrapper'],
        },

        test_suite='setup.my_test_suite',
        install_requires=[
            'numpy',
'scipy',
'numba',
'pandas',
'networkx',
'tqdm',

        ]
    )


# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from tempfile import NamedTemporaryFile
import moleculekit.home
import numpy
import os
import ctypes as ct


def xtc_lib():
    lib = {}
    libdir = moleculekit.home.home(libDir=True)

    import platform
    if platform.system() == "Windows":
        lib['libc'] = ct.cdll.msvcrt
        ct.cdll.LoadLibrary(os.path.join(libdir, "libgcc_s_seh-1.dll"))
        if os.path.exists(os.path.join(libdir, "psprolib.dll")):
            ct.cdll.LoadLibrary(os.path.join(libdir, "psprolib.dll"))
        lib['libxtc'] = ct.cdll.LoadLibrary(os.path.join(libdir, "libxtc.dll"))
    else:
        # lib['libc'] = cdll.LoadLibrary("libc.so.6")
        lib['libc'] = ct.cdll.LoadLibrary("libc.{}".format("so.6" if platform.uname()[0] != "Darwin" else "dylib"))
        lib['libxtc'] = ct.cdll.LoadLibrary(os.path.join(libdir, "libxtc.so"))
    return lib


def string_to_tempfile(content, ext):
    f = NamedTemporaryFile(delete=False, suffix="." + ext)
    f.write(content.encode("ascii"))
    f.close()
    return f.name


def pack_string_buffer(data):
    '''data1 = data.astype(dtype=numpy.string_)

#    buf = [create_string_buffer(data[i].encode('ascii')) for i in range(len(data))]
    buf = [create_string_buffer(data1[i]) for i in range(len(data1))]
    ptr = (c_char_p * len(data1))()

    for i in range(0, len(data1)):
        ptr[i] = addressof(buf[i])'''

    # Stefan alternative
    ptr = (ct.c_char_p * len(data))()
    ptr[:] = data.astype(dtype=numpy.string_).tolist()
    #ptr[:] = [x.encode() for x in data]

    return ptr


def pack_int_buffer(data):
    ptr = (ct.c_int * len(data))()
    ptr[:] = data
    '''i = 0
    for d in data:
        ptr[i] = d
        i += 1'''
    return ptr


def pack_ulong_buffer(data):
    ptr = (ct.c_ulong * len(data))()
    ptr[:] = data
    '''i = 0
    for d in data:
        ptr[i] = d
        i += 1'''
    return ptr


def pack_double_buffer(data):
    ptr = (ct.c_double * len(data))()
    ptr[:] = data
    '''i = 0
    for d in data:
        ptr[i] = d
        i += 1'''
    return ptr



from rdkit.Chem import AllChem
from rdkit.Chem.rdForceFieldHelpers import UFFGetMoleculeForceField


def restrainedEmbed(mol, atomPos={}, restraintStrengths=(100, 25), forceField=UFFGetMoleculeForceField, 
                    numMinim=40, **kwargs):
    """ Restrained embedding of a molecule
    
    Generates an embedding of a molecule where part of the molecule is restrained to specific coordinates.
    It works in an iterative manner minimizing with different restraint strength in each iteration.

    Parameters
    ----------
    mol : rdkit.Chem.Mol
        The molecule to embed. This function will modify the molecule in place.
    atomPos : dict
        A dictionary with key the index of the atom as an integer and value a list of the three xyz coordinates
        where the atom should be restrained to. For the atoms for which no keys are in the dictionary, they will
        be left unrestrained.
    restraintStrengths : list
        A list of restraints strengths which will be applied in order.
    forceField : rdkit ForceField
        The rdkit forcefield to use for the minimization.
    numMinim : int
        The number of minimizations to perform at each iteration.
    """

    for restr in restraintStrengths:
        ff = forceField(mol, confId=0)
        conf = mol.GetConformer()
        for i in range(mol.GetNumAtoms()):
            p = conf.GetAtomPosition(i)
            if i in atomPos:
                p.x, p.y, p.z = float(atomPos[i][0]), float(atomPos[i][1]), float(atomPos[i][2])
                pIdx = ff.AddExtraPoint(p.x, p.y, p.z, fixed=True) - 1
                ff.AddDistanceConstraint(pIdx, i, 0, 0, float(restr))

        ff.Initialize()
        n = numMinim
        more = ff.Minimize(energyTol=1e-4, forceTol=1e-3)
        while more and n:
            more = ff.Minimize(energyTol=1e-4, forceTol=1e-3)
            n -= 1
